#include "mpi.h"
#include "mpiimpl.h"

int MPI_Spawn2() {
  int sd, ns;                /* socket descriptors            */
  struct init_msg msg1;
  int fromlen, rank;
  struct sockaddr_in new_sin;   /* used for each new connection   */

  //Leave the loader alone if no space in host table
  if (num_slaves == PG_ARRAY_SIZE) return -1;

  //Start up new slave process
  _is_spawn2 = 1;
  is_initialized = 0;
  do_unexec();
  is_initialized = 1;
  _is_spawn2 = 0;
  if (!attach_new_slaves()) return -1;

  //Allocate new slave record. Easy
  rank = ++num_slaves;

  //Accept connection from new slave, remember it
  fromlen = sizeof(new_sin);
  CALL_CHK( ns = accept, (_mpi_my_list_sd, (struct sockaddr *)&new_sin, &fromlen) );
  pg_array[rank].sd = ns;
  FD_SET(ns, &fdset);
  if ( ns > max_sd )
    max_sd = ns;
#ifdef DEBUG
  printf("master:  accepted new slave connection\n");
  printf("master:  slave %d: new ns:  %d, current max_sd: %d\n", rank, ns, max_sd);
#endif

  //Receive slave's listener_addr and store in field of pg_array[rank]
  CALL_CHK( recv, (ns, (char *)&(pg_array[rank].listener_addr),
            sizeof(struct sockaddr_in), 0) );

  //Send back number of slaves and the assigned rank
  msg1.len = htonl(sizeof(msg1));
  msg1.rank = htonl(rank);
  msg1.num_slaves = htonl(num_slaves);
  send(ns, (char *)&msg1, sizeof(msg1), 0);

  //Send listener_addr fields to slave
  { char buf[PG_ARRAY_SIZE * sizeof(struct sockaddr_in)], *ptr;
    int i;

    ptr = buf;
    for ( i=0; i <= num_slaves; i++ ) {
      memcpy( ptr, (char *)&(pg_array[i].listener_addr),
	           sizeof(struct sockaddr_in) );
      ptr += sizeof(struct sockaddr_in);
    }
    send( pg_array[rank].sd, buf, (1+num_slaves)*sizeof(struct sockaddr_in), 0 );

    //Acknowledge that master and slave are synchronized
    *((INT *)buf) = htonl( sizeof(struct sockaddr_in) );
    send( pg_array[rank].sd, buf, sizeof(INT), 0);
  }
}
